#!/usr/bin/env python3
from BioMetaDB.DBOperations.fix import fix
from BioMetaDB.Accessories.arg_parse import ArgParse
from BioMetaDB.DBOperations.move_project import move_project
from BioMetaDB.Accessories.program_caller import ProgramCaller
from BioMetaDB.DBOperations.integrity_check import integrity_check
from BioMetaDB.DBOperations.create_database import create_database
from BioMetaDB.DBOperations.delete_from_table import delete_from_table
from BioMetaDB.DBOperations.summarize_database import summarize_database
from BioMetaDB.DBOperations.update_existing_table import update_existing_table
from BioMetaDB.Exceptions.record_list_exceptions import ColumnNameNotFoundError
from BioMetaDB.Exceptions.config_manager_exceptions import TableNameNotFoundError, MultipleTablesFoundError, \
    ConfigFileNotFound
from BioMetaDB.DBOperations.remove_columns_from_table import remove_columns_from_table
from BioMetaDB.DBOperations.remove_table_from_database import remove_table_from_database
from BioMetaDB.Exceptions.remove_columns_from_table_exceptions import ListFileNotProvidedError
from BioMetaDB.DBOperations.create_table_in_existing_database import create_table_in_existing_database


def run():
    args_list = (
        (("program",),
         {"help": "Program to run"}),
        (("-n", "--db_name"),
         {"help": "Name of database"}),
        (("-t", "--table_name"),
         {"help": "Name of database table", "default": "None"}),
        (("-d", "--directory_name"),
         {"help": "Directory path with bio data (fasta or fastq)", "default": "None"}),
        (("-f", "--data_file"),
         {"help": ".tsv or .csv file to add, or .fix file to integrate", "default": "None"}),
        (("-l", "--list_file"),
         {"help": "File with list of items, typically ids or column names", "default": "None"}),
        (("-c", "--config_file"),
         {"help": "/path/to/BioMetaDB-project-directory - Can omit if no other BioMetaDB project present"}),
        (("-a", "--alias"),
         {"help": "Provide alias for locating and creating table class", "default": "None"}),
        (("-s", "--silent"),
         {"help": "Silence all standard output (Standard error still displays to screen)", "action": "store_true",
          "default": False}),
        (("-v", "--view"),
         {"help": "View (c)olumns or (t)ables with SUMMARIZE", "default": "None"}),
        (("-q", "--query"),
         {"help": "evaluation ~> genome; function -> gen; eval >> fxn; eval ~> fxn -> gen;", "default": "None"}),
        (("-u", "--unique"),
         {"help": "View unique values of column using SUMMARIZE", "default": "None"}),
        (("-i", "--integrity_cancel"),
         {"help": "Cancel integrity check", "default": False, "action": "store_true"}),
        (("-w", "--write"),
         {"help": "Write fastx records from SUMMARIZE to outfile", "default": "None"}),
        (("-x", "--write_tsv"),
         {"help": "Write table record metadata from SUMMARIZE to outfile", "default": "None"}),
        (("-p", "--path"),
         {"help": "New path for moving project in MOVE command", "default": "None"}),
        (("-r", "--truncate"),
         {"help": "Return only ID of annotation and not the complete description, excluding Prokka, default False",
          "default": False, "action": "store_true"}),
    )
    programs = {
        "INIT":                     create_database,
        "UPDATE":                   update_existing_table,
        "CREATE":                   create_table_in_existing_database,
        "REMOVECOL":                remove_columns_from_table,
        "DELETE":                   delete_from_table,
        "REMOVE":                   remove_table_from_database,
        "SUMMARIZE":                summarize_database,
        "MOVE":                     move_project,
        "INTEGRITY":                integrity_check,
        "FIX":                      fix,
    }
    flags = {
        "INIT":                 ("db_name", "table_name", "directory_name", "data_file", "alias", "silent", "integrity_cancel"),
        "CREATE":               ("config_file", "table_name", "directory_name", "data_file", "alias", "silent", "integrity_cancel"),
        "UPDATE":               ("config_file", "table_name", "directory_name", "data_file", "alias", "silent", "integrity_cancel"),
        "REMOVECOL":            ("config_file", "table_name", "list_file", "alias", "silent", "integrity_cancel"),
        "DELETE":               ("config_file", "table_name", "list_file", "alias", "silent", "integrity_cancel"),
        "REMOVE":               ("config_file", "table_name", "alias", "silent", "integrity_cancel"),
        "SUMMARIZE":            ("config_file", "view", "query", "table_name", "alias", "write", "write_tsv", "unique", "truncate"),
        "MOVE":                 ("config_file", "path", "integrity_cancel", "silent"),
        "INTEGRITY":            ("config_file", "table_name", "alias", "silent"),
        "FIX":                  ("config_file", "data_file", "silent", "integrity_cancel"),
    }
    errors = {
        TableNameNotFoundError:     "Name of table not found",
        ListFileNotProvidedError:   "List file not provided",
        ColumnNameNotFoundError:    "Column name not in table",
        MultipleTablesFoundError:   "Multiple projects located - specify with -c /path/to/BioMetaDB-project-directory",
        ConfigFileNotFound:         "Config file for project not found",
    }
    _help = {
        "INIT":             "Initialize database with starting table, fasta directory, and/or data files",
        "CREATE":           "Create a new table in an existing database, optionally populate using data files",
        "UPDATE":           "Update values in existing table or add new sequences",
        "REMOVECOL":        "Remove column list (including data) from table",
        "DELETE":           "Delete list of ids from database tables, remove associated files",
        "REMOVE":           "Remove table and all associated data from database",
        "SUMMARIZE":        "Summarize project and query data. Write records or metadata to file",
        "MOVE":             "Move project to new location",
        "INTEGRITY":        "Queries project database and structure to generate .fix file for possible issues",
        "FIX":              "Repairs errors in DB structure using .fix file",
    }

    ap = ArgParse(args_list,
                  description=ArgParse.description_builder("dbdm:\tManage BioMetaDB project",
                                                           _help, flags))

    pc = ProgramCaller(programs=programs, flags=flags, _help=_help, errors=errors)

    pc.run(ap.args, debug=True)


if __name__ == "__main__":
    run()



"""
Class holds extensions used by indexers
Carries dictionary for parsing indexed fasta files

"""


class IndexExtensions:
    IDX_FILE = ".imidx"
    IDX_FA_FILE = ".imfadx"
    IDX_FQ_FILE = ".imfqdx"
    IDX_TSV_FILE = ".imtbx"
    match = {
        IDX_FA_FILE:    "fasta",
        IDX_FQ_FILE:    "fastq",
        "fasta":        IDX_FA_FILE,
        "fastq":        IDX_FQ_FILE,
    }



from sqlalchemy.orm import mapper
from BioMetaDB.Models.models import BaseData
from BioMetaDB.Models.functions import Record
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.DataStructures.record_list import RecordList
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.Config.config_manager import ConfigManager, Config
from BioMetaDB.Accessories.update_data import UpdateData
from BioMetaDB.Serializers.tsv_joiner import TSVJoiner
from .dbdm import run


def get_table(config_path, table_name=None, alias=None):
    """ Primary method for users importing

    :param config_path:
    :param table_name:
    :param alias:
    :return:
    """
    # Load config data from file
    cfg, config_path = ConfigManager.confirm_config_set(config_path)
    if alias:
        table_name = ConfigManager.get_name_by_alias(alias, cfg)
    config = ConfigManager(cfg, table_name)
    # Get session to return
    engine = BaseData.get_engine(config.db_dir, config.db_name + ".db")
    sess = BaseData.get_session_from_engine(engine)
    # Generate table class and name based on presence of alias
    TableClass = ClassManager.get_class_orm(table_name, engine)
    UserClass = type(table_name, (Record,), {})
    # Map to SQL orm
    mapper(UserClass, TableClass)
    return RecordList(sess, UserClass, config)


def tables(config_path):
    """ Primary method for listing tables in project

    :param config_path:
    :return:
    """
    cfg, config_path = ConfigManager.confirm_config_set(config_path)
    return list(cfg[ConfigKeys.TABLES_TO_DB].keys())


def genomes(config_path):
    return [val for val in tables(config_path) if val not in ("functions", "evaluation")]


if __name__ == "__main__":
    run()



import os
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, create_session
from sqlalchemy.ext.declarative import declarative_base

""" Class for handling Bsse schema requests, gathering metadata and session info, etc.



"""


class BaseData:
    """ Class for storing database config values, initial creation, and creating session

    """
    Base = declarative_base()

    @staticmethod
    def get_uri(basedir, db_name):
        """

        :param db_name: (str)	Name of database, preferably from child class DBNames
        """
        return os.environ.get('DATABASE_URL') or 'sqlite:///' + os.path.join(basedir, db_name)

    @staticmethod
    def create(basedir, db_name):
        """ Creates all tables in engine

        :param basedir: (str) Base directory of database folder
        :param db_name: (str) Name of database to create. Should reference from child class DBNames
        """
        SQLALCHEMY_DATABASE_URI = BaseData.get_uri(basedir, db_name)
        engine = create_engine(SQLALCHEMY_DATABASE_URI)
        # Creates all tables in database
        BaseData.Base.metadata.create_all(engine)

    @staticmethod
    def get_session(basedir, db_name):
        """ Returns session variable

        :param basedir: (str) Base directory of database folder
        :param db_name: (str) Name of database for which to create session
        :return Object:
        """
        SQLALCHEMY_DATABASE_URI = BaseData.get_uri(basedir, db_name)
        engine = create_engine(SQLALCHEMY_DATABASE_URI)
        # Returns session for CRUD operations on database
        return sessionmaker(bind=engine)()

    @staticmethod
    def get_session_from_engine(e):
        """ Returns session variable

        :param e: (Engine)  Engine object
        :return Object:
        """
        # Returns session for CRUD operations on database
        return create_session(bind=e, autocommit=False, autoflush=True)

    @staticmethod
    def get_engine(basedir, db_name):
        """ Returns session variable

        :param basedir: (str) Base directory of database folder
        :param db_name: (str) Name of database for which to create session
        :return Object:
        """
        SQLALCHEMY_DATABASE_URI = BaseData.get_uri(basedir, db_name)
        engine = create_engine(SQLALCHEMY_DATABASE_URI)
        # Returns session for CRUD operations on database
        return engine



import os
from BioMetaDB.Config.directory_manager import Directories
from BioMetaDB.Exceptions.config_manager_exceptions import ConfigAssertString
from BioMetaDB.Serializers.count_table import CountTable
from BioMetaDB.Config.config_manager import ConfigManager
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.DBManagers.type_mapper import TypeMapper
from BioMetaDB.Exceptions.create_database_exceptions import CreateDBAssertString
from BioMetaDB.DBOperations.integrity_check import integrity_check

"""
Script will hold functionality for CREATE, to create new tables when in existing database
Function called from ProgramCaller

"""


def _create_table_display_message_prelude(db_name, working_directory, table_name, directory_name, data_file, alias):
    """ Display summary of input parameters before creating database

    :param db_name: (str)   Name of db
    :param working_directory: (str) Path to working directory
    :param table_name: (str)    Record that will be created
    :param directory_name: (str)    Directory with files to add
    :param data_file: (str)     File with metadata for storing in database
    :return:
    """
    print("CREATE:\tCreate table in existing database")
    print(" Project root directory:\t%s" % working_directory)
    print(" Name of database:\t\t%s.db" % db_name.strip(".db"))
    print(" Name of table:\t\t\t%s" % table_name)
    print(" Table aliases:\t\t\t%s" % alias, "\n")
    print("DATA:\tPopulate table")
    print(" Get metadata from\t\t%s" % data_file)
    print(" Copy fastx files from\t\t%s" % directory_name, "\n")


def _create_table_display_message_epilogue():
    print("New table creation complete!", "\n")


def _create_all_directories(working_directory, table_name):
    """ Creates directories for table and class info

    :param working_directory: (str) Name of package directory
    :param table_name: (str)    Name of db table to create
    :return:
    """
    db_dir = os.path.join(working_directory, Directories.DATABASE)
    os.makedirs(os.path.join(db_dir, table_name), exist_ok=True)


def create_table_in_existing_database(config_file, table_name, directory_name, data_file, alias, silent, integrity_cancel):
    """

    :param silent:
    :param config_file:
    :param table_name:
    :param directory_name:
    :param data_file:
    :param alias:
    :return:
    """
    assert table_name != "None", CreateDBAssertString.TABLE_NAME_NOT_SET
    assert config_file is not None, ConfigAssertString.CONFIG_FILE_NOT_PASSED
    _cfg, _tbl, _sil, _al = config_file, table_name, silent, alias
    config, config_file = ConfigManager.confirm_config_set(config_file)
    if table_name in config.keys():
        print("!! Record exists, exiting. To update table, use UPDATE !!")
        exit(1)
    if not silent:
        _create_table_display_message_prelude(config[ConfigKeys.DATABASES][ConfigKeys.db_name],
                                              config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir],
                                              table_name,
                                              directory_name,
                                              data_file,
                                              alias)
    data_types = {}
    initial_data = []
    if data_file is not "None":
        _initial_data = CountTable(data_file)
        # Ignore first column name, typically announcing "Name" or "Genome ID"
        # Get names and types of each column in dict
        # Key is header name (which will be used as name of column in database)
        # Value is type of data (int, float, str) that is found for a random genome id's data value
        # TODO: Data type determination requires a uniformity from the .tsv file data. Consider a workaround
        data_types = TypeMapper.get_translated_types(_initial_data, TypeMapper.py_type_to_string)
        initial_data.append(_initial_data)
    # Gather bio data from folder
    if directory_name != "None":
        genomic_files_to_add = (_f for _f in os.listdir(directory_name))
    else:
        genomic_files_to_add = ()
    # Create new table directories
    _create_all_directories(config[ConfigKeys.DATABASES][ConfigKeys.working_dir], table_name)
    # Update config object with new data
    table_name = table_name.lower()
    config[table_name] = {
        ConfigKeys.rel_classes_dir: os.path.join(config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir],
                                                 Directories.CLASSES),
    }
    config.set(ConfigKeys.TABLES_TO_DB, table_name, config[ConfigKeys.DATABASES][ConfigKeys.db_name])
    config.set(ConfigKeys.TABLES_TO_ALIAS, "{}|{}".format(alias, table_name), table_name)
    # Write new config file
    with open(config_file, "w") as W:
        config.write(W)
    # Update ConfigManager object
    cfg = ConfigManager(config, table_name)
    # Create new table and populate with new data
    ClassManager.create_initial_table_in_db(cfg.db_name, cfg.working_dir, table_name, data_types, silent, initial=False)
    if data_file is not "None":
        for _data in initial_data:
            ClassManager.populate_data_to_existing_table(table_name, _data, cfg,
                                                         genomic_files_to_add, directory_name, silent)
    if not silent:
        _create_table_display_message_epilogue()
    if not integrity_cancel:
        integrity_check(_cfg, _tbl, _al, _sil)



class CreateDBAssertString:
    WORKING_DIR_EXISTS = "Directory exists, exiting"
    WORKING_DB_NOT_SET = "Database name not set"
    SEQUENCE_DIR_NOT_EXISTS = "Sequence directory does not exist, exiting"
    TABLE_NAME_NOT_SET = "Record name not set"



#!/usr/bin/env python3

import os
import shutil
import sqlite3
import sqlalchemy
from Models.models import BaseData
from examples.planctomycetes import Planctomycetes
from Accessories.arg_parse import ArgParse
from Accessories.program_caller import ProgramCaller, db_validate

"""
Script handles database management operations

"""


@db_validate
def add_file(class_name, file_path, seq_type, *args, **kwargs):
    """ Primary method for adding a file not located in db directory to the directory
    Stored metadata to database

    :param class_name: (str)	Name of database class
    :param file_path: (str)	file to add to database
    :param seq_type: (str)	Determines location for moving file, referenced from .FileLocations of class
    """
    # Get table class as class and create database session
    table_class, sess = get_class(class_name)
    # Get file metadata from name
    _, path_and_file = os.path.splitdrive(file_path)
    _, file = os.path.split(path_and_file)
    file = file.split(".")
    # Correct datatype based on file extension
    if file[-1] == "fasta" or file[-1] == "fa":
        file[-1] = "fna"
    elif file[-1] == "fastq":
        file[-1] = "fq"
    elif "protein" in file:
        file[-1] = "faa"
    # Instantiate object using id from file name, location based on passed class,
    # and subfolder based on seq_type
    gen_id = ".".join(file[:-1])
    db_object = table_class(genome_id=gen_id, location=getattr(table_class.FileLocations, seq_type), data_type=file[-1])
    # Add to database and copy file to new path
    if input(
            "Add {} to database {} in {}? ".format(gen_id, table_class.__tablename__, table_class.db_name)).upper() in (
    "Y", "YES"):
        sess.add(db_object)
        shutil.copy(path_and_file, db_object.full_path())
        sess.commit()
        exit(0)


@db_validate
def remove_file(class_name, genome_id):
    """ Primary method for removing file and metadata from db

    :param class_name: (str)	Name of class to get from dictionary
    :param genome_id: (str)		Genome ID as stored in database
    """
    table_class, sess = get_class(class_name)
    if input("Are you sure you want to delete this file?\nWARNING: THIS CANNOT BE UNDONE: ").upper() in ("Y", "YES"):
        db_object = sess.query(table_class).filter(table_class.genome_id == genome_id).first()
        if db_object:
            os.remove(db_object.full_path())
            sess.delete(db_object)
            sess.commit()
            print("{} successfully deleted from {}!".format(genome_id, class_name))
        else:
            print("{} does not exist in {} within {}".format(genome_id, table_class.__tablename__, table_class.db_name))


def get_class(class_name):
    """ Returns class based on dictionary mapping and instantiated session object

    :param class_name: (str)	User-passed name of class
    :return Tuple[class, dbObject]:
    """
    table_class = classes[class_name]
    sess = BaseData.get_session(table_class.db_name)
    return table_class, sess


if __name__ == "__main__":
    args_list = [
        [["program"],
         {"help": "Select program"}],
        [["-f", "--file_path"],
         {"help": ".faa, .fna, or .fq file"}],
        [["-c", "--class_name"],
         {"help": "Name of Record"}],
        [["-s", "--seq_type"],
         {"help": "Sequence type (example GEN)"}],
        [["-g", "--genome_id"],
         {"help": "Genome ID"}]
    ]

    ap = ArgParse(args_list,
                  description="dbdm:\tProgram for managing database operations.\n\nAvailable Programs:\n\nADD: Add "
                              "file to database and move to directory\n\t(Req: -f, -c, -s)\n\nREMOVE: Remove sequence "
                              "from database and delete from directory\n\t(Req: -c, -g)")

    classes = {
        "Planctomycetes": Planctomycetes,
        "Plancto": Planctomycetes,
    }

    programs = {
        "ADD": add_file,
        "REMOVE": remove_file,
    }

    flags = {
        "ADD": ("file_path", "seq_type", "class_name"),
        "REMOVE": ("genome_id", "class_name")
    }

    errors = {
        IOError: "Error moving file to location, rolling back changes",
        AttributeError: "Database name not found",
        KeyError: "Record name not found",
        sqlite3.IntegrityError: "Sequence exists in database matching file's genome id, exiting",
        sqlalchemy.exc.IntegrityError: "Sequence exists in database matching file's genome id, exiting",
    }

    _help = {
        "ADD": "Adds file to database and copy file to database folder",
        "REMOVE": "Remove file from database and remove from database folder",
    }

    pc = ProgramCaller(flags=flags, programs=programs, classes=classes, errors=errors, _help=_help)

    pc.run(ap.args)



from BioMetaDB.Exceptions.error import Error


class TableNameNotFoundError(Error):
    pass


class MultipleTablesFoundError(Error):
    pass


class ConfigFileNotFound(Error):
    pass


class ConfigAssertString:
    CONFIG_FILE_NOT_PASSED = "Config file for project not passed"
    CONFIG_FILE_NOT_FOUND = "Config file for project not found"


class TableNameAssertString:
    TABLE_NOT_FOUND = "Record not found"



import os
from BioMetaDB.Accessories.ops import touch
from BioMetaDB.Accessories.ops import print_if_not_silent
from BioMetaDB.Serializers.count_table import CountTable
from BioMetaDB.Config.directory_manager import Directories
from BioMetaDB.Exceptions.create_database_exceptions import CreateDBAssertString
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.DBManagers.type_mapper import TypeMapper
from BioMetaDB.Config.config_manager import Config
from BioMetaDB.Config.config_manager import ConfigManager
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.DBOperations.integrity_check import integrity_check


"""
Function handles creating database and first table
Initializes with provided directory and data file

"""


def _initialization_display_message_prelude(db_name, working_directory, table_name, directory_name, data_file, alias):
    """ Display summary of input parameters before creating database

    :param db_name: (str)   Name of db
    :param working_directory: (str) Path to working directory
    :param table_name: (str)    Record that will be created
    :param directory_name: (str)    Directory with files to add
    :param data_file: (str)     File with metadata for storing in database
    :return:
    """
    print("INIT:\tCreate project database and initialize")
    print(" Project root directory:\t%s" % working_directory)
    print(" Name of database:\t\t%s.db" % db_name.strip(".db"))
    print(" Name of table:\t\t\t%s" % table_name)
    print(" Table aliases:\t\t\t%s" % alias, "\n")
    print("DATA:\tPopulate table")
    print(" Get metadata from\t\t%s" % data_file)
    print(" Copy fastx files from\t\t%s" % directory_name, "\n")


def _initialization_display_message_epilogue():
    print("Initialization complete!", "\n")


def _create_all_directories(working_directory, table_name):
    """ Creates directories for package, database, config, and class info

    :param working_directory: (str) Name of package directory
    :param table_name: (str)    Name of db table to create
    :return:
    """
    os.makedirs(working_directory)
    classes_dir = os.path.join(working_directory, Directories.CLASSES)
    config_dir = os.path.join(working_directory, Directories.CONFIG)
    db_dir = os.path.join(working_directory, Directories.DATABASE)
    migrations_dir = os.path.join(working_directory, Directories.MIGRATIONS)
    os.makedirs(classes_dir)
    os.makedirs(config_dir)
    os.makedirs(db_dir)
    os.makedirs(migrations_dir)
    table_dir = os.path.join(db_dir, table_name)
    return classes_dir, config_dir, db_dir, table_dir


def create_database(db_name, table_name, directory_name, data_file, alias, silent, integrity_cancel):
    """ Function called from dbdm initializes project/module

    :param silent:
    :param alias:
    :param db_name: (str)   Name of db
    :param table_name: (str)    Record that will be created
    :param directory_name: (str)    Directory with files to add
    :param data_file: (str)     File with metadata for storing in database
    :return:
    """
    # Confirm working dir does not exist and that directory with genomes does exist
    assert db_name != "None", CreateDBAssertString.WORKING_DB_NOT_SET
    assert table_name != "None", CreateDBAssertString.TABLE_NAME_NOT_SET
    assert os.path.isdir(db_name) is False, CreateDBAssertString.WORKING_DIR_EXISTS
    if directory_name != "None":
        assert os.path.isdir(directory_name), CreateDBAssertString.SEQUENCE_DIR_NOT_EXISTS
    table_name = table_name.lower()
    if not silent:
        _initialization_display_message_prelude(db_name, db_name, table_name, directory_name, data_file, alias)
    # Gather files to commit and initial data to store for each file
    print_if_not_silent(silent, "Beginning process...")
    print_if_not_silent(silent, " Getting files from %s" % directory_name)
    if directory_name != "None":
        genomic_files_to_add = (_f for _f in os.listdir(directory_name))
    else:
        genomic_files_to_add = ()
    data_types = {}
    initial_data = []
    if data_file is not "None":
        _initial_data = CountTable(data_file)
        # Ignore first column name, typically announcing "Name" or "Genome ID"
        # Get names and types of each column in dict
        # Key is header name (which will be used as name of column in database)
        # Value is type of data (int, float, str) that is found for a random genome id's data value
        # TODO: Data type determination requires a uniformity from the .tsv file data. Consider a workaround
        data_types = TypeMapper.get_translated_types(_initial_data, TypeMapper.py_type_to_string)
        initial_data.append(_initial_data)
    # Create working directories
    print_if_not_silent(silent, " Creating directories at database root %s" % db_name)
    classes_dir, config_dir, db_dir, table_dir = _create_all_directories(db_name, table_name)
    # Create database file
    print_if_not_silent(silent, " Creating database file in %s" % db_dir)
    touch(os.path.join(db_dir, db_name + ".db"))
    # Write configuration info
    config_file = db_name + ".ini"
    print_if_not_silent(silent, " Writing database configuration to %s" % os.path.join(config_dir, config_file))
    config = Config()
    abs_path_working_dir = os.path.abspath(db_name)
    db_name = os.path.basename(db_name)
    config[ConfigKeys.DATABASES] = {
        ConfigKeys.db_name: db_name,
        ConfigKeys.working_dir: abs_path_working_dir,
        ConfigKeys.rel_work_dir: db_name,
        ConfigKeys.migrations_dir: os.path.join(abs_path_working_dir, Directories.MIGRATIONS),
        ConfigKeys.config_dir: os.path.join(abs_path_working_dir, Directories.CONFIG),
        ConfigKeys.db_dir: os.path.join(abs_path_working_dir, Directories.DATABASE),
        ConfigKeys.rel_db_dir: os.path.join(db_name, Directories.DATABASE),
    }
    config[table_name] = {
        ConfigKeys.rel_classes_dir:     os.path.join(db_name, Directories.CLASSES),
    }
    config[ConfigKeys.TABLES_TO_DB] = {
        table_name: db_name,
    }
    config[ConfigKeys.TABLES_TO_ALIAS] = {
        "{}|{}".format(alias, table_name): table_name,
    }
    with open(os.path.join(config_dir, config_file), "w") as W:
        config.write(W)
    # Create table
    print_if_not_silent(silent, "Creating new table %s at %s" % (table_name, os.path.join(db_dir, db_name)))
    os.makedirs(table_dir)
    ClassManager.create_initial_table_in_db(db_name, db_name, table_name, data_types, silent, initial=False)
    # Populate table with data from file and genomes
    # Get config file - confirms that it was written correctly
    cfg = ConfigManager(config, table_name)
    if data_file is not "None":
        for _data in initial_data:
            ClassManager.populate_data_to_existing_table(table_name, _data, cfg,
                                                         genomic_files_to_add, directory_name, silent)
    if not silent:
        _initialization_display_message_epilogue()
    if not integrity_cancel:
        integrity_check(db_name, table_name, "None", silent)



import os
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.Config.config_manager import ConfigManager
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.Serializers.count_table import CountTable
from BioMetaDB.DBOperations.integrity_check import integrity_check

"""
Script holds functionality for UPDATE, which updates existing/adds new records to database
Handles new column generation and SQL migration 

"""


def _update_display_message_prelude(db_name, working_directory, table_name, directory_name, data_file, alias):
    """ Display summary of input parameters before creating database

    :param db_name: (str)   Name of db
    :param working_directory: (str) Path to working directory
    :param table_name: (str)    Record that will be created
    :param directory_name: (str)    Directory with files to add
    :param data_file: (str)     File with metadata for storing in database
    :return:
    """
    print("UPDATE:\tUpdate table in existing database")
    print(" Project root directory:\t%s" % working_directory)
    print(" Name of database:\t\t%s.db" % db_name.strip(".db"))
    print(" Name of table:\t\t\t%s" % table_name)
    print(" Table aliases:\t\t\t%s" % alias, "\n")
    print("DATA:\tPopulate table")
    print(" Get metadata from\t\t%s" % data_file)
    print(" Copy fastx files from\t\t%s" % directory_name, "\n")


def _update_table_display_message_epilogue():
    print("Update complete!", "\n")


def update_existing_table(config_file, table_name, directory_name, data_file, alias, silent, integrity_cancel):
    """

    :param alias:
    :param silent:
    :param config_file:
    :param table_name:
    :param directory_name:
    :param data_file:
    :return:
    """
    _cfg, _tbl, _al, _sil = config_file, table_name, alias, silent
    config, config_file = ConfigManager.confirm_config_set(config_file)
    if alias != "None":
        table_name = ConfigManager.get_name_by_alias(alias, config)
        if table_name is None:
            print(
                "!! Table does not exist! Run CREATE to add to existing database, or INIT to create in new database !!"
            )
            exit(1)
    if table_name != "None" and table_name not in config.keys():
        print("!! Table does not exist! Run CREATE to add to existing database, or INIT to create in new database !!")
        exit(1)
    cfg = ConfigManager(config, table_name)
    if not silent:
        _update_display_message_prelude(
            config[ConfigKeys.DATABASES][ConfigKeys.db_name],
            config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir],
            table_name,
            directory_name,
            data_file,
            alias
        )
    if data_file != "None":
        data_to_add = CountTable(data_file)
    else:
        data_to_add = None
    if directory_name != "None":
        genomic_files_to_add = (_f for _f in os.listdir(directory_name))
    else:
        genomic_files_to_add = ()
    if alias == "None":
        new_attrs = ClassManager.populate_data_to_existing_table(table_name, data_to_add, cfg, genomic_files_to_add,
                                                                 directory_name, silent, alias)
    else:
        new_attrs = ClassManager.populate_data_to_existing_table(table_name, data_to_add, cfg, genomic_files_to_add,
                                                                 directory_name, silent)
    ClassManager.write_class(new_attrs, cfg.classes_file)
    if not silent:
        _update_table_display_message_epilogue()
    if not integrity_cancel:
        integrity_check(_cfg, _tbl, _al, _sil)



import operator
import random
from sqlalchemy import Float, Integer, String, VARCHAR, Boolean

from BioMetaDB.Serializers.count_table import CountTable

"""
Class holds dictionaries to map types to each other and to string representations

"""


class TypeMapper:
    py_type_to_string = {
        float:      "Float",
        int:        "Integer",
        str:        "String",
        bool:       "Boolean",
    }
    sql_type_to_string = {
        Float:      "Float",
        Integer:    "Integer",
        String:     "String",
        Boolean:    "Boolean",
    }
    string_to_sql_type = {
        "Float": Float,
        "Integer": Integer,
        "String": String,
        "Boolean": Boolean,
    }
    string_to_loaded_sql_type = {
        "Float": Float,
        "Integer": Integer,
        "String": VARCHAR,
        "Boolean": Boolean,
    }
    string_to_py_type = {
        "Float": float,
        "Integer": int,
        "String": str,
        "Boolean": bool,
    }
    py_type_to_sql_type = {
        float:      Float,
        int:        Integer,
        str:        String,
        bool:       Boolean,
    }
    sql_type_to_py_type = {
        Float:      float,
        Integer:    int,
        String:     str,
        Boolean:    bool,
    }
    py_bool_to_sql_bool = {
        "True": "TRUE",
        "False": "FALSE",
    }

    @staticmethod
    def get_translated_types(counttable_object, dict_to_reference):
        if dict_to_reference == TypeMapper.py_type_to_string:
            return {
                counttable_object.header[i]:
                    dict_to_reference[TypeMapper.determine_column_type(counttable_object, i)]
                for i in range(len(counttable_object.header))
            }
        return {
            counttable_object.header[i]:
                dict_to_reference[
                    type(counttable_object.get_at(random.sample(counttable_object.file_contents.keys(), 1)[0], i))
                ]
            for i in range(len(counttable_object.header))
        }

    @staticmethod
    def determine_column_type(counttable_object, column_idx):
        values = {
            int: 0,
            bool: 0,
            float: 0,
            str: 0,
            # None: 0
        }
        for value_list in counttable_object.file_contents.values():
            if value_list[column_idx] != 'None':
                values[
                    CountTable._try_return_type(value_list[column_idx])
                ] += 1
        return (max(values.items(), key=lambda x: x[1]) if values.items() else (str,))[0]



# cython: language_level=3
import os
from sqlalchemy.orm import mapper
from random import randint
from datetime import datetime
from BioMetaDB.Models.models import BaseData
from BioMetaDB.Models.functions import Record
from BioMetaDB.Config.config_manager import ConfigManager, ConfigKeys
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.DBManagers.integrity_manager import IntegrityManager

"""
Script will hold functionality that allows user to fix integrity issues within the project structure
These issues include:
    

"""


def _integrity_check_display_message_prelude(db_name, working_directory, tables_to_search, fixfile_prefix):
    """ Display initial message summarizing operation on files

    :param db_name:
    :param working_directory:
    :param fixfile_prefix:
    :param tables_to_search:
    :return:
    """
    print("INTEGRITY:\tCheck project for data issues")
    print(" Project root directory:\t%s" % working_directory)
    print(" Name of database:\t\t%s" % db_name)
    print(" Tables in check:\t\t%s" % ", ".join(tables_to_search).rstrip(", "))
    print(" Fix file generated:\t\t%s" % fixfile_prefix)
    print(" Note: .fix file will auto-delete if no issues are found", "\n")


def _integrity_check_display_message_epilogue(passed, fix_file):
    if passed == 0:
        print("Integrity check complete with no errors!", "\n")
    else:
        print("Integrity check complete, see %s for errors!" % fix_file, "\n")


def integrity_check(config_file, table_name, alias, silent):
    """  Function called from dbdm that checks integrity and issues in project at all levels

    :param config_file:
    :param table_name:
    :param alias:l
    :return:
    """
    config, config_file = ConfigManager.confirm_config_set(config_file)
    if alias != "None":
        table_name = ConfigManager.get_name_by_alias(alias, config)
    if table_name is None or (table_name == "None" and alias == "None"):
        tables_to_search = list(config[ConfigKeys.TABLES_TO_DB].keys())
    else:
        tables_to_search = [table_name, ]
    py_fixfile_name = "%s.%s.fix" % (datetime.today().strftime("%Y%m%d"),
                                     str(randint(1, 1001)))
    if not silent:
        _integrity_check_display_message_prelude(config[ConfigKeys.DATABASES][ConfigKeys.db_name],
                                                 config[ConfigKeys.DATABASES][ConfigKeys.working_dir],
                                                 tables_to_search,
                                                 py_fixfile_name)
    im = IntegrityManager(config, py_fixfile_name, tables_to_search)
    # TODO: Implement table-level checks
    im.initialize_fix_file()
    im.initial_project_check()
    # im.table_check()
    for table in tables_to_search:
        cfg = ConfigManager(config, table)
        engine = BaseData.get_engine(cfg.db_dir, cfg.db_name + ".db")
        sess = BaseData.get_session_from_engine(engine)
        TableClass = ClassManager.get_class_orm(table, engine)
        UserClass = type(table, (Record,), {})
        # Map to SQL orm
        mapper(UserClass, TableClass)
        im.record_check(sess, UserClass, table)
    if im.issues_found == 0:
        os.remove(py_fixfile_name)
    if not silent:
        _integrity_check_display_message_epilogue(im.issues_found, py_fixfile_name)
    del im



import os
import random
import shutil
from Bio import SeqIO
from random import randint
from string import ascii_letters
from BioMetaDB.Accessories.ops import chunk
from BioMetaDB.Accessories.bio_ops import BioOps
from BioMetaDB.Accessories.ops import read_until
from BioMetaDB.Indexers.index_extensions import IndexExtensions
from BioMetaDB.Exceptions.get_exceptions import SequenceIdNotFoundError
from BioMetaDB.Exceptions.get_exceptions import ImproperFormatIndexFileError


class IndexMapper:
    def __init__(self, file_name):
        """

        :param file_name: (str)	Load data into dictionaries
        """
        if IndexExtensions.IDX_FILE not in file_name:
            file_name += IndexExtensions.IDX_FILE
        self.data_dict, self.inv_data_dict = IndexMapper._load_file(file_name)

    @staticmethod
    def is_indexed(file_name):
        """ Checks if index file exists

        :param file_name: (str) Name of file
        :return bool:
        """
        return os.path.isfile(os.path.join(file_name, IndexExtensions.IDX_FILE))

    @staticmethod
    def _load_file(file_name, CHUNK_SIZE=100000):
        """ Loads file data into two of dictionaries

        :param file_name: (str)	Name of file to map
        :return Tuple[Dict[str, str], Dict[str, str]]:
        """
        data_dict = {}
        inv_data_dict = {}
        with open(file_name, "rb") as R:
            # Load large files by chunks
            for lines in chunk(R, CHUNK_SIZE):
                for line in lines:
                    line = line.decode().rstrip("\r\n").split("\t")
                    if len(line) == 3:
                        data_dict[line[0]] = (line[2], int(line[1]))
                        inv_data_dict[line[2]] = (line[0], int(line[1]))
                    elif len(line) == 2:
                        data_dict[line[0]] = (line[1],)
                        inv_data_dict[line[1]] = (line[0],)
                    else:
                        raise ImproperFormatIndexFileError
        return data_dict, inv_data_dict

    def get(self, key):
        """ Retrieve associated match based on either key or value

        :param key: (str)	Value to retrieve, can either be key or value
        :return str:
        """
        try:
            return self.data_dict[key][0]
        except KeyError:
            try:
                return self.inv_data_dict[key][0]
            except KeyError as e:
                raise SequenceIdNotFoundError(e)

    def get_line_num(self, key):
        """ Retrieve associated match for line number based on either key or value

        :param key: (str)	Value to retrieve, can either be key or value
        :return int:
        """
        assert len(self.data_dict[key]) == 2, "File does not store line numbers"
        try:
            return self.data_dict[key][1]
        except KeyError:
            try:
                return self.inv_data_dict[key][1]
            except KeyError as e:
                raise SequenceIdNotFoundError(e)


class IndexCreator:

    @staticmethod
    def make_idx_for_directory_file_names(directory_name):
        """ Make index file for filenames in a directory

        :param directory_name: (str)    Name of directory with files to rename
        :return:
        """
        current_dir = os.getcwd()
        W = open(os.path.join(current_dir, os.path.dirname(directory_name) + IndexExtensions.IDX_FILE), "wb")
        files_in_directory = set(file for file in os.listdir(directory_name))
        for _file in files_in_directory:
            data_type = BioOps.get_type(_file)
            W.write(("%s\t%s\n" % (_file, IndexCreator._line_edit(16) + "." + data_type)).encode())
        W.close()

    @staticmethod
    def create_from_file(file_name, CHUNK_SIZE=100000):
        """ Create shortened id index from list of genomes

        :param CHUNK_SIZE: (int)    Chunk size for fast reading
        :param file_name: (str)	    Name of file for which to create index
        """
        line_count = 0
        W = open(file_name + IndexExtensions.IDX_FILE, "wb")
        with open(file_name, "rb") as R:
            # Load large files by chunks
            for lines in chunk(R, CHUNK_SIZE):
                for line in lines:
                    line_count += 1
                    line = line.decode().rstrip("\r\n")
                    W.write(("%s\t%s\t%s\n" % (line, str(line_count), IndexCreator._line_edit(16))).encode())
        W.close()

    @staticmethod
    def create_from_fastx(file_name, CHUNK_SIZE=100000):
        """ Create index of short names for a fastx file

        :param file_name: (str)	Name of fastx file
        :param CHUNK_SIZE: (int)    Chunk size for fast reading
        """
        line_count = 0
        # Read file by bytes
        W = open(file_name + IndexExtensions.IDX_FILE, "wb")
        with open(file_name, "rb") as R:
            # Load large files by chunks
            for lines in chunk(R, CHUNK_SIZE):
                for line in lines:
                    line = line.decode().rstrip("\r\n")
                    line_count += 1
                    if line.startswith(">") or line.startswith("@"):
                        W.write(("%s\t%s\t%s\n" % (
                            read_until(line[1:], " "), str(line_count), IndexCreator._line_edit(16)
                        )).encode())
        W.close()

    @staticmethod
    def rewrite_ids_in_fastx(image_mapper, file_name, remove_description=True):
        """ Rename fastx ids using index file, write to new file

        :param remove_description: (bool)   Only write ids in new file
        :param image_mapper: (IndexMapper)	Mapped index file
        :param file_name: (str)	Name of fastx file
        """
        data_type = BioOps.get_type(file_name)
        records = BioOps.parse_large(file_name, data_type)
        for record in records:
            record.id = image_mapper.get(record.id)
            if remove_description:
                record.description = ""
        SeqIO.write(records, file_name + IndexExtensions.match[data_type], data_type)

    @staticmethod
    def rename_directory_contents_using_index(directory):
        """ Uses index file to rename all files in a folder
        Note that index file name must match directory name

        :param directory: (str) Directory with files to rename
        :return:
        """
        im = IndexMapper(directory)
        for _file in os.listdir(directory):
            try:
                shutil.move(os.path.join(directory, _file), os.path.join(directory, im.get(_file)))
            except SequenceIdNotFoundError:
                print("Could not move %s" % _file)

    @staticmethod
    def _line_edit(new_length):
        """ Get shortened version of line

        :param new_length: (int)    Length of id to create
        :return str:
        """
        shortened_line = ""
        for i in range(new_length - 5):
            shortened_line += random.choice(ascii_letters)
        shortened_line += str(randint(10000, 99999))
        return shortened_line



class FixAssertString:
    FIX_NOT_FOUND = "Fix file not found or missing .fix extension"
    BAD_PROJECT_DIR = "Project directory passed does not exist"
    INTERNAL_BAD_KWARGS = "INTERNAL ERROR: Bad **kwargs passed"



"""
Class for use by user - Build a DataTable and use this to update records and to write to a tsv

"""
import os


class Data:
    """ Data node for building objects with key:value attributes

    """
    def __init__(self, _id=""):
        self._id = _id
        self.data = {}

    def _get(self):
        """ Returns a dictionary consisting of the item and its attributes

        :return:
        """
        return self.data

    def get(self):
        return self._id, self.data

    def delattr(self, attr):
        """ Removes column of info storage

        :param attr:
        :return:
        """
        if attr in self.data.keys():
            del self.data[attr]
            return True
        return False

    def __getattr__(self, attr):
        """

        :return:
        """
        try:
            return self.data[attr]
        except KeyError:
            raise AttributeError

    def setattr(self, key, value):
        """ Function sets key, value pair for this object

        :param key:
        :param value:
        :return:
        """
        self.data[key] = value


class UpdateData:

    def __init__(self):
        self.tsv = ""
        self.num_records = 0
        self._header = set()
        self.data = [Data(), ]
        self._ids = {}

    def __repr__(self):
        """ Calls get function for string representation

        :return:
        """
        return str(self.get())

    def __delattr__(self, item):
        """ Builtin function support

        :param item:
        :return:
        """
        self.delcol(item)

    def add(self, _id):
        """ Method for adding a new id to the list. Automatically called when UpdateData is accessed for
        a non-existing id

        :param _id:
        :return:
        """
        if self.num_records == 0:
            self.data[0] = Data(_id=_id)
            self.num_records = 1
            self._ids[_id] = 0
            return
        self.data.append(Data(_id=_id))
        self.num_records += 1
        self._ids[_id] = self.num_records - 1

    def delcol(self, col_name):
        """ Removes tracked column from UpdateData structure tracking

        :param col_name:
        :return:
        """
        range_obj = range(0, len(self.data))
        success = False
        for i in range_obj:
            success = self.data[i].delattr(col_name)
        if not success:
            raise KeyError("Column name not found")

    def keys(self):
        """ Returns the ids of all Data objects stored in the UpdateData object

        :return:
        """
        for value in self.data:
            for val in value.data.keys():
                self._header.add(val)
        return self._header

    def get(self):
        """ Return list of dictionaries, each composed of a Data object

        :return:
        """
        return [{val._id: val._get()} for val in self.data]

    def __getitem__(self, item):
        """ Allows class to be indexed and searched

        :param item:
        :return:
        """
        # Index of stored list
        # ID stored
        item_type = type(item)
        if item_type == str:
            # Return matching record
            _id = self._ids.get(item, None)
            if _id is not None:
                return self.data[_id]
            # Not found, add to data
            self.add(item)
            # Get newest added value
            return self.data[-1]
        elif item_type == int:
            assert item < self.num_records, "Index must be less than length"
            return self.data[item]
        # Slice of indices
        elif item_type == slice:
            if type(item.start) is None:
                item.start = 0
            if type(item.stop) is None:
                item.stop = -1
            if type(item.step) is None:
                item.start = 1
            return tuple(self.data[i] for i in range(item.start, item.stop, item.step))
        raise TypeError("Unable to determine type")

    def __add__(self, other):
        assert type(other) == UpdateData, "Must combine two UpdateData objects"
        for item in other.data:
            _id = self._ids.get(item, None)
            if _id is not None:
                for key, val in item._get().items():
                    self.data[_id].setattr(key, val)
            else:
                self.data.append(Data(_id=item._id))
                self.num_records += 1
                for key, val in item._get().items():
                    self.data[-1].setattr(key, val)
                self._ids[item._id] = self.num_records - 1
        return self

    def __delitem__(self, item):
        """ Delete item from list of stored data

        :param item:
        :return:
        """
        if type(item) == int:
            assert item < self.num_records, "Index must be less than length"
            del self.data[item]
            self.num_records -= 1
        # Slice of indices
        elif type(item) == slice:
            if type(item.start) is None:
                item.start = 0
            if type(item.stop) is None:
                item.stop = -1
            if type(item.step) is None:
                item.start = 1
            for i in range(item.start, item.stop, item.step):
                del self.data[i]
                self.num_records -= 1
        # ID stored
        elif type(item) == str:
            # Delete matching record
            _id = self._ids.get(item, None)
            if _id is not None:
                del self.data[_id]
                self.num_records -= 1
                return
            raise ValueError("Item id not found")
        raise TypeError("Unable to determine type")

    def to_file(self, file_name, delim="\t", na_rep="None", _order=None, skip_header=False):
        """ Write entire results to tsv or csv file, filling in gaps as needed with na_rep
        _order is list to follow to output header
        Option to skip header in output

        :param file_name:
        :param na_rep:
        :param delim:
        :param _order:
        :param skip_header:
        :return:
        """
        # Store name of file
        self.tsv = UpdateData._handle_filename(file_name)
        W = open(self.tsv, "w")
        # Write tsv header
        if _order is None:
            header = list(self.keys())
        else:
            header = list(_order)
        header_line = "ID\t"
        if not skip_header:
            for head in header:
                header_line += head + "\t"
            W.write(header_line[:-1] + "\n")
        # Write data, filling in as needed
        for data in self.data:
            W.write(data._id)
            for head in header:
                val = data.data.get(head, None)
                if val:
                    W.write(delim + str(val))
                else:
                    W.write(delim + na_rep)
            W.write("\n")
        W.close()

    def delete_file(self):
        """ Removes tsv file stored as object attribute

        :return:
        """
        if os.path.exists(self.tsv):
            os.remove(self.tsv)

    def __len__(self):
        """ Accessible through len()

        :return:
        """
        return self.num_records

    @staticmethod
    def _handle_filename(file_name):
        """ Protected method to handle loading ~ and .. in filenames

        :param file_name:
        :return:
        """
        if "~" in file_name:
            return os.path.relpath(os.path.expanduser(file_name))
        return os.path.relpath(file_name)

    @staticmethod
    def from_file(file_name, has_header=True, delim="\t", na_rep="None", skip_lines=0, comment_delim="#"):
        """ Read in tsv/csv file into UpdateData object. Can add to existing object

        :param file_name:
        :param has_header:
        :param delim:
        :param na_rep:
        :param skip_lines:
        :param comment_delim:
        :param initial_data:
        :return:
        """
        initial_data = UpdateData()
        assert type(na_rep) == str, "Parameter `na_rep` must be of type str"
        assert skip_lines >= 0, "Number of initial lines to skip must be an integer greater than or equal to 0"
        R = open(UpdateData._handle_filename(file_name), "r")
        for i in range(skip_lines):
            next(R)
        if has_header:
            header = next(R).rstrip("\r\n").split(delim)
        else:
            header = "Column"
        # Write data, filling in as needed
        for _line in R:
            # Skip over comment lines, if provided
            if comment_delim is not None and _line.startswith(comment_delim):
                continue
            line = _line.rstrip("\r\n").split(delim)
            line_len = len(line)
            for i in range(1, line_len):
                if line[i] == na_rep:
                    line[i] = None
                if has_header:
                    initial_data[line[0]].setattr(header[i], line[i])
                else:
                    initial_data[line[0]].setattr(header + str(i), line[i])
        return initial_data



import os
import shutil
from sqlalchemy.orm import mapper
from BioMetaDB.DBOperations.integrity_check import integrity_check
from BioMetaDB.Exceptions.config_manager_exceptions import TableNameAssertString
from BioMetaDB.Models.models import BaseData
from BioMetaDB.Models.functions import Record
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.Accessories.ops import print_if_not_silent
from BioMetaDB.Config.config_manager import ConfigManager
from BioMetaDB.DBManagers.class_manager import ClassManager

"""
Script will hold functionality for REMOVE, which removes a table from a database,
deleting all associated files

"""


def _remove_table_display_message_prelude(db_name, working_directory, table_name, alias):
    """ Display summary of input parameters before creating database

    :param alias:
    :param db_name: (str)   Name of db
    :param working_directory: (str) Path to working directory
    :param table_name: (str)    Record that will be created
    :return:
    """
    print("REMOVE:\tCreate table in existing database")
    print(" Project root directory:\t%s" % working_directory)
    print(" Name of database:\t\t%s.db" % db_name.strip(".db"))
    print(" Name of table:\t\t\t%s" % table_name)
    print(" Table aliases:\t\t\t%s" % alias, "\n")
    print("DATA:\tDelete table")
    print(" Record name:\t%s" % table_name, "\n")


def _remove_columns_display_message_epilogue():
    print("Record removed from database!", "\n")


def remove_table_from_database(config_file, table_name, alias, silent, integrity_cancel):
    """ Function removes a given table from a database

    :param silent:
    :param config_file:
    :param table_name:
    :param alias:
    :return:
    """
    _cfg, _tbl, _sil, _al = config_file, table_name, silent, alias
    config, config_file = ConfigManager.confirm_config_set(config_file)
    if alias != "None":
        table_name = ConfigManager.get_name_by_alias(alias, config)
        assert table_name is not None, TableNameAssertString.TABLE_NOT_FOUND
    if not silent:
        _remove_table_display_message_prelude(
            config[ConfigKeys.DATABASES][ConfigKeys.db_name],
            config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir],
            table_name,
            alias
        )

    cfg = ConfigManager(config, table_name)
    engine = BaseData.get_engine(cfg.db_dir, cfg.db_name + ".db")
    sess = BaseData.get_session_from_engine(engine)
    TableClass = ClassManager.get_class_orm(table_name, engine)
    UserClass = type(table_name, (Record,), {})
    mapper(UserClass, TableClass)
    all_records = sess.query(UserClass).all()
    for record in all_records:
        if record:
            print_if_not_silent(silent, " ..Removing record %s" % record._id)
            try:
                os.remove(record.full_path())
            except OSError:
                continue
    TableClass.drop(engine)
    cfg.remove_table_from_config_file(table_name)
    os.remove(cfg.classes_file)
    shutil.rmtree(cfg.table_dir)
    if not silent:
        _remove_columns_display_message_epilogue()
    if not integrity_cancel:
        integrity_check(_cfg, _tbl, _al, _sil)



from sqlalchemy.orm import mapper
from BioMetaDB.Models.models import BaseData
from BioMetaDB.Models.functions import Record
from BioMetaDB.DataStructures.record_list import RecordList
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.Config.config_manager import ConfigManager, ConfigKeys
from BioMetaDB.Exceptions.summarize_database_exceptions import SummarizeDBAssertString


"""
Script will hold functionality to display to stdout a summary of key project details

"""


def _summarize_display_message_prelude(db_name):
    """ Display summary of input parameters before displaying db summary

    :param db_name: (str)   Name of db
    :return:
    """
    print("SUMMARIZE:\tView summary of all tables in database")
    print(" Project root directory:\t%s" % db_name)
    print(" Name of database:\t\t%s.db\n" % db_name.strip(".db"))


def summarize_database(config_file, view, query, table_name, alias, write, write_tsv, unique, truncate):
    """ Function will query all tables listed in the config file, outputting simple
    metrics to the screen

    :param config_file:
    :param view:
    :param query:
    :param table_name:
    :param alias:
    :param write:
    :param write_tsv:
    :param truncate:
    :return:
    """
    if not view:
        if query != "None":
            assert (query != "None" and (table_name != "None" or alias != "None")), SummarizeDBAssertString.QUERY_AND_TABLE_SET
        assert not (table_name != "None" and alias != "None"), SummarizeDBAssertString.ALIAS_OR_TABLE_ONLY
    config, config_file = ConfigManager.confirm_config_set(config_file)
    if not view.lower()[0] == "t":
        _summarize_display_message_prelude(config[ConfigKeys.DATABASES][ConfigKeys.db_name])
    if alias != "None":
        table_name = ConfigManager.get_name_by_alias(alias, config)
    if table_name != "None" or alias != "None":
        tables_in_database = (table_name, )
    else:
        tables_in_database = config[ConfigKeys.TABLES_TO_DB].keys()
    if ("~>" in query) and ("->" in query):
        assert table_name == 'None', "Query cannot contain '~>/->' statement with a table name"
        matching_records = []
        eval_sess, EvalClass, eval_cfg = load_table_metadata(config, "evaluation")
        eval_rl = RecordList(eval_sess, EvalClass, eval_cfg, compute_metadata=False, truncate=truncate)
        fxn_sess, FxnClass, fxn_cfg = load_table_metadata(config, "functions")
        fxn_rl = RecordList(fxn_sess, FxnClass, fxn_cfg, compute_metadata=False, truncate=truncate)
        evaluation_query, block_1 = query.split("~>")
        function_query, annotation_query = block_1.split("->")
        if evaluation_query.replace(" ", "") != '':
            eval_rl.query(evaluation_query)
        else:
            eval_rl.query()
        if function_query.replace(" ", "") != '':
            fxn_rl.query(function_query)
        else:
            fxn_rl.query()
        for record in eval_rl:
            if record in fxn_rl:
                record_id = record._id.split(".")[0]
                sess, UserClass, cfg = load_table_metadata(config, record_id)
                annot_rl = RecordList(sess, UserClass, cfg, compute_metadata=False, truncate=truncate)
                _handle_query(annot_rl, annotation_query)
                if write_tsv != 'None':
                    annot_rl.write_tsv(record_id + "." + write_tsv.replace(".tsv", "") + ".tsv")
                for record_2 in annot_rl:
                    matching_records.append(record_2)
                if write_tsv == 'None' and write == 'None':
                    annot_rl.summarize()
        if matching_records and write != "None":
            rl = RecordList(compute_metadata=False, records_list=matching_records, truncate=truncate)
            rl.write_records(write)
        return
    if "~>" in query:
        assert table_name == 'None', "Query cannot contain a '~>' statement with a table name"
        matching_records = []
        eval_sess, EvalClass, eval_cfg = load_table_metadata(config, "evaluation")
        eval_rl = RecordList(eval_sess, EvalClass, eval_cfg, compute_metadata=False, truncate=truncate)
        evaluation_query, annotation_query = query.split("~>")
        if evaluation_query.replace(" ", "") != '':
            eval_rl.query(evaluation_query)
        else:
            eval_rl.query()
        for record in eval_rl:
            record_id = record._id.split(".")[0]
            sess, UserClass, cfg = load_table_metadata(config, record_id)
            annot_rl = RecordList(sess, UserClass, cfg, compute_metadata=False, truncate=truncate)
            _handle_query(annot_rl, annotation_query)
            if write_tsv != 'None':
                annot_rl.write_tsv(record_id + "." + write_tsv.replace(".tsv", "") + ".tsv")
            for record_2 in annot_rl:
                matching_records.append(record_2)
            if write_tsv == 'None' and write == 'None':
                annot_rl.summarize()
        if matching_records and write != "None":
            rl = RecordList(compute_metadata=False, records_list=matching_records, truncate=truncate)
            rl.write_records(write)
        return
    if "->" in query:
        assert table_name == 'None', "Query cannot contain a '->' statement with a table name"
        matching_records = []
        eval_sess, EvalClass, eval_cfg = load_table_metadata(config, "functions")
        eval_rl = RecordList(eval_sess, EvalClass, eval_cfg, compute_metadata=False, truncate=truncate)
        evaluation_query, annotation_query = query.split("->")
        if evaluation_query.replace(" ", "") != '':
            eval_rl.query(evaluation_query)
        else:
            eval_rl.query()
        for record in eval_rl:
            record_id = record._id.split(".")[0]
            sess, UserClass, cfg = load_table_metadata(config, record_id)
            annot_rl = RecordList(sess, UserClass, cfg, compute_metadata=False, truncate=truncate)
            _handle_query(annot_rl, annotation_query)
            if write_tsv != 'None':
                annot_rl.write_tsv(record_id + "." + write_tsv.replace(".tsv", "") + ".tsv")
            for record_2 in annot_rl:
                matching_records.append(record_2)
            if write_tsv == 'None' and write == 'None':
                annot_rl.summarize()
        if matching_records and write != "None":
            rl = RecordList(compute_metadata=False, records_list=matching_records, truncate=truncate)
            rl.write_records(write)
        return
    if ">>" in query:
        assert table_name == 'None', "Query cannot contain a '>>' statement with a table name"
        eval_sess, EvalClass, eval_cfg = load_table_metadata(config, "evaluation")
        eval_rl = RecordList(eval_sess, EvalClass, eval_cfg, compute_metadata=False, truncate=truncate)
        evaluation_query, annotation_query = query.split(">>")
        if evaluation_query.replace(" ", "") != '':
            eval_rl.query(evaluation_query)
        else:
            eval_rl.query()
        sess, UserClass, cfg = load_table_metadata(config, "functions")
        if len(eval_rl) > 0:
            in_query = ""
            for record in eval_rl:
                in_query += "_id == '%s' OR " % record._id
            annot_rl = RecordList(sess, UserClass, cfg, compute_metadata=False, truncate=truncate)
            if annotation_query.replace(" ", "") != '':
                annot_rl.query(annotation_query + " AND " + in_query[:-4])
            else:
                annot_rl.query(in_query[:-4])
            if annot_rl is not None:
                if write != "None":
                    annot_rl.write_records(write)
                if write_tsv != "None":
                    annot_rl.write_tsv(write_tsv.replace(".tsv", "") + ".tsv")
                if write_tsv == 'None' and write == 'None':
                    annot_rl.summarize()
        else:
            eval_rl.summarize()
        return
    if view == "None" and unique == 'None':
        for tbl_name in tables_in_database:
            # Display queried info for single table and break
            if table_name == 'None' or table_name == tbl_name:
                sess, UserClass, cfg = load_table_metadata(config, tbl_name)
                annot_rl = RecordList(sess, UserClass, cfg, compute_metadata=True, truncate=truncate)
                _handle_query(annot_rl, query)
                if len(annot_rl) > 1:
                    annot_rl.summarize()
                elif len(annot_rl) == 1:
                    print(annot_rl[0])
    for tbl_name in tables_in_database:
        sess, UserClass, cfg = load_table_metadata(config, tbl_name)
        annot_rl = RecordList(sess, UserClass, cfg, truncate=truncate)
        # Display column info for table
        if view.lower()[0] == "c":
            # Display queried info for single table and break
            # Do not need to query since only displaying columns
            annot_rl.columns_summary()
        elif view.lower()[0] == "t":
            # Display queried info for single table and break
            # Do not need to query since only displaying columns
            annot_rl.table_name_summary()
        if write != "None" and table_name == tbl_name:
            _handle_query(annot_rl, query)
            annot_rl.write_records(write)
        if write_tsv != "None" and table_name == tbl_name:
            _handle_query(annot_rl, query)
            annot_rl.write_tsv(write_tsv)
        if unique != 'None' and table_name == tbl_name:
            _handle_query(annot_rl)
            col_vals = set()
            for record in annot_rl:
                val = getattr(record, unique, "None")
                if val:
                    col_vals.add(val)
            col_vals = sorted(col_vals)
            for val in col_vals:
                print(val)


def load_table_metadata(config, tbl_name):
    cfg = ConfigManager(config, tbl_name)
    engine = BaseData.get_engine(cfg.db_dir, cfg.db_name + ".db")
    sess = BaseData.get_session_from_engine(engine)
    TableClass = ClassManager.get_class_orm(tbl_name, engine)
    UserClass = type(tbl_name, (Record,), {})
    # Map to SQL orm
    mapper(UserClass, TableClass)
    # Display queried info for single table and break
    return sess, UserClass, cfg


def _handle_query(rl, query="None"):
    if query != "None":
        rl.query(query)
    else:
        rl.query()



from datetime import datetime
from BioMetaDB.DBOperations.integrity_check import integrity_check
from BioMetaDB.Exceptions.config_manager_exceptions import TableNameAssertString
from BioMetaDB.Models.models import BaseData
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.Config.config_manager import ConfigManager
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.DBManagers.update_manager import UpdateManager
from BioMetaDB.Exceptions.remove_columns_from_table_exceptions import ListFileNotProvidedError

"""
Script will hold functionality for REMOVECOL, to remove columns from existing table data
Saves migration
Creates new database table model

"""


def _remove_columns_display_message_prelude(db_name, working_directory, table_name, alias, cols_to_remove):
    """ Display initial message summarizing operation on files

    :param db_name:
    :param working_directory:
    :param table_name:
    :param alias:
    :param cols_to_remove:
    :return:
    """
    print("REMOVECOL:\tRemove list of columns from database schema")
    print(" Project root directory:\t%s" % working_directory)
    print(" Name of database:\t%s" % db_name)
    print(" Name of table:\t\t%s" % table_name)
    print(" Table aliases:\t\t\t%s" % alias, "\n")
    print("DATA:\tRemove columns from schema")
    print(" Do not include columns:\t%s" % ",".join(cols_to_remove).strip(","), "\n")


def _remove_columns_display_message_epilogue():
    print("Columns removed from database!", "\n")


def remove_columns_from_table(config_file, table_name, list_file, alias, silent, integrity_cancel):
    """

    :param silent:
    :param config_file:
    :param table_name:
    :param list_file:
    :param alias:
    :return:
    """
    _cfg, _tbl, _sil, _al = config_file, table_name, silent, alias
    config, config_file = ConfigManager.confirm_config_set(config_file)
    if alias != "None":
        table_name = ConfigManager.get_name_by_alias(alias, config)
        assert table_name is not None, TableNameAssertString.TABLE_NOT_FOUND
    if list_file == "None":
        raise ListFileNotProvidedError
    cfg = ConfigManager(config, table_name)
    columns_to_remove = set(line.rstrip("\r\n") for line in open(list_file, "r"))
    if not silent:
        _remove_columns_display_message_prelude(
            config[ConfigKeys.DATABASES][ConfigKeys.db_name],
            config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir],
            table_name,
            alias,
            columns_to_remove
        )

    engine = BaseData.get_engine(cfg.db_dir, cfg.db_name + ".db")
    sess = BaseData.get_session_from_engine(engine)
    TableClass = ClassManager.get_class_orm(table_name, engine)
    update_manager = UpdateManager(cfg, ClassManager.get_class_as_dict(cfg), sess)
    table_copy_csv = update_manager.create_table_copy(datetime.today().strftime("%Y%m%d"), TableClass, silent)
    new_attrs = {key: value
                 for key, value
                 in ClassManager.correct_dict(ClassManager.get_class_as_dict(cfg)).items()
                 if key not in columns_to_remove}
    UpdatedDBClass, metadata = ClassManager.generate_class(
        cfg.table_name,
        new_attrs,
        cfg.db_dir,
        cfg.db_name,
        cfg.table_dir
    )
    ClassManager.write_class(new_attrs, cfg.classes_file)
    cfg.update_config_file(table_name)
    update_manager.delete_old_table_and_populate(
        engine,
        TableClass,
        UpdatedDBClass,
        table_copy_csv,
        table_name,
        sess,
        silent
    )
    if not silent:
        _remove_columns_display_message_epilogue()
    if not integrity_cancel:
        integrity_check(_cfg, _tbl, _al, _sil)



#!/usr/bin/env python3
from BioMetaDB.Accessories.arg_parse import ArgParse
from BioMetaDB.Accessories.ops import chunk
import fileinput
import sys
import os


CHUNK_SIZE = 1000


def issue_match(search_params_set, issue_info, replace_dict):
    """ Function checks if user-set criteria in search_params was met by the particular issue
    data in *args

    :return:
    """
    # Issue matches search criteria passed by user
    set_values = set(issue_info.values())
    for key in search_params_set:
        if key not in set_values:
            return None
    replace_string = list(replace_dict.values())[0]
    for item in issue_info.keys():
        if item is not None and issue_info[item] is not None:
            replace_string = replace_string.replace("<%s>" % item, issue_info[item])
    return replace_string


if __name__ == "__main__":
    args_list = (
        (("-m", "--match"),
         {"help": "Look for issues matching list, ex: RECORD,BAD_LOCATION,genomic", "required": True}),
        (("-r", "--replace"),
         {"help": "Set value using list, ex: FILE,/path/to/<ID>.fna", "required": True}),
        (("-f", "--fix_file"),
         {"help": "/path/to/generated_fix_file.fix", "required": True}),
    )

    ap = ArgParse(args_list, description="Script will make changes to .fix file\n")

    R = fileinput.FileInput(ap.args.fix_file, inplace=True, backup=".bak")
    search_params = set([val for val in ap.args.match.split(",") if val != ""])
    _replace_lookup = [_v for _v in ap.args.replace.split(",") if _v != ""]
    replace_lookup = {_replace_lookup[0]: _replace_lookup[1]}
    # For possibly large files
    _id = ""
    data_type = ""
    location = None
    fix_data = None
    issue_type = ""
    fix_type = ""
    replace_id = None
    fix_data_col = list(replace_lookup.keys())[0]
    try:
        for batch in chunk(R, CHUNK_SIZE):
            for line in R:
                while line.startswith("@"):
                    line = next(R)
                    sys.stdout.write(line)
                while line == "ISSUE:\n":
                    # Read in '<DATA_TYPE:\t<ID>[:\t<LOCATION]\n' line
                    line = next(R)
                    sys.stdout.write(line)
                    line = line.rstrip("\r\n").split(":\t")
                    data_type = line[0]
                    _id = line[1]
                    if len(line) == 3:
                        location = line[2]
                    # Read in '<ISSUE_TYPE>\t<FIX_TYPE>[\t<FIX_DATA>]\n' line
                    line = next(R)
                    line = line.rstrip("\r\n").split("\t")
                    issue_type = line[0]
                    fix_type = line[1]
                    fix_data = issue_match(
                        search_params,
                        {"DATA_TYPE": data_type,
                         "ID": _id,
                         "LOCATION": location,
                         "FIX_DATA": fix_data,
                         "ISSUE_TYPE": issue_type,
                         "FIX_TYPE": fix_type},
                        replace_lookup
                    )
                    if fix_data:
                        sys.stdout.write("%s\t%s\t%s\n" % (line[0], fix_data_col, fix_data))
                    else:
                        sys.stdout.write("%s\t%s\n" % (line[0],"\t".join(line[1:])))
                    # Read in '---' line
                    line = next(R)
                    sys.stdout.write(line)
                    # Read in next issue header
                    line = next(R)
                    sys.stdout.write(line)
                    fix_data = None
    except StopIteration:
        pass
    R.close()
    os.remove(ap.args.fix_file + ".bak")



from operator import itemgetter

"""
Functions handle simple range operations
Primarily works with data in the form ((int, int),)
for given ((start, end),) coordinates

"""


def reduce_span(coords_list):
    """ Function will reduce ranges into tuple of non-overlapping ranges

    :param coords_list: [(int, int),]    Tuple/List of start/end coordinates
    :return ((int,int),):
    """
    # Sort coordinates by start value
    ranges_in_coords = sorted(coords_list, key=itemgetter(0))
    # Will group together matching sections into spans
    # Return list of these spans at end
    # Initialize current span and list to return
    spans_in_coords = [list(ranges_in_coords[0]), ]
    for coords in ranges_in_coords[1:]:
        # The start value is within the farthest range of current span
        # and the end value extends past the current span
        if coords[0] <= spans_in_coords[-1][1] < coords[1]:
            spans_in_coords[-1][1] = coords[1]
        # The start value is past the range of the current span
        # Append old span to list to return
        # Reset current span to this new range
        elif coords[0] > spans_in_coords[-1][1]:
            spans_in_coords.append(list(coords))
    return tuple(map(tuple, spans_in_coords))


def gaps(coords_list, gap_size=1):
    """ Function will locate uncovered portions between ranges

    :param gap_size: (int)  Return gaps of size or larger
    :param coords_list: [(int, int),]   Tuple/List of start/end coordinates
    :return ((int, int),):
    """
    # Reduce list of coordinates to ranges
    coords_span = reduce_span(coords_list)
    # Return gaps in between ranges if they are at least the gap size
    return tuple((coords_span[i][1] + 1, coords_span[i + 1][0] - 1) for i in range(len(coords_span) - 1)
                 # Add 1 to difference to include single-unit gaps
                 if (coords_span[i + 1][0] - 1) - (coords_span[i][1] + 1) + 1 >= gap_size)



import os
from Cython.Build import cythonize
import setuptools
from setuptools import setup, Extension
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), 'README.md'), "r") as f:
    long_description = f.read()

# with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "requirements.txt"), "r") as fp:
#     install_requires = fp.read().strip().split("\n")

extensions = [
    Extension("BioMetaDB.Accessories.arg_parse", ["BioMetaDB/Accessories/arg_parse.c"]),
    Extension("BioMetaDB.Accessories.bio_ops", ["BioMetaDB/Accessories/bio_ops.c"]),
    Extension("BioMetaDB.Accessories.ops", ["BioMetaDB/Accessories/ops.c"]),
    Extension("BioMetaDB.Accessories.program_caller", ["BioMetaDB/Accessories/program_caller.c"]),
    Extension("BioMetaDB.DataStructures.record_list", ["BioMetaDB/DataStructures/record_list.c"]),
    Extension("BioMetaDB.DBManagers.class_manager", ["BioMetaDB/DBManagers/class_manager.c"]),
    Extension("BioMetaDB.DBManagers.integrity_manager", ["BioMetaDB/DBManagers/integrity_manager.c"]),
    Extension("BioMetaDB.DBManagers.update_manager", ["BioMetaDB/DBManagers/update_manager.c"]),
    Extension("BioMetaDB.Models.functions", ["BioMetaDB/Models/functions.c"]),
    Extension("BioMetaDB.Serializers.count_table", ["BioMetaDB/Serializers/count_table.c"]),
    Extension("BioMetaDB.Serializers.fix_file", ["BioMetaDB/Serializers/fix_file.c"]),
    Extension("BioMetaDB.Serializers.tsv_joiner", ["BioMetaDB/Serializers/tsv_joiner.c"]),
]

setup(
    name='BioMetaDB',
    version='0.1.2.31',
    description='Use biological data to generate SQL database schema',
    url="https://github.com/cjneely10/BioMetaDB",
    author="Christopher Neely",
    author_email="christopher.neely1200@gmail.com",
    license="GNU GPL 3",
    install_requires=[
        "SQLAlchemy==1.3.12",
        "biopython==1.76",
        "configparser==3.8.1",
        "Cython==0.29.14"
    ],
    python_requires='>=3.6',
    ext_modules=extensions,
    packages=setuptools.find_packages(),
    entry_points={
        'console_scripts': [
            'dbdm = BioMetaDB.dbdm:run'
        ]
    },
)




import os
from BioMetaDB.Config.config_manager import ConfigManager, ConfigKeys
from BioMetaDB.Exceptions.fix_exceptions import FixAssertString
from BioMetaDB.DBManagers.integrity_manager import IntegrityManager
from BioMetaDB.DBOperations.integrity_check import integrity_check


def _fix_display_message_prelude(db_name, working_directory, fixfile_prefix):
    """ Display initial message summarizing operation on files

    :param db_name:
    :param working_directory:
    :param fixfile_prefix:
    :return:
    """
    print("FIX:\tUse .fix file to repair project issues")
    print(" Project root directory:\t%s" % working_directory)
    print(" Name of database:\t\t%s" % db_name)
    print(" Fix file to use:\t\t%s\n" % fixfile_prefix)


def _fix_display_message_epilogue():
    print("\nProject fix complete! Re-run INTEGRITY to confirm!", "\n")


def fix(config_file, data_file, silent, integrity_cancel):
    """ Function called from dbdm that commits all fixes listed in .fix file to project

    :param config_file:
    :param data_file:
    :param silent:
    :return:
    """
    _cfg, _sil = config_file, silent
    config, config_file = ConfigManager.confirm_config_set(config_file)
    assert os.path.exists(data_file) and os.path.splitext(data_file)[1] == ".fix", FixAssertString.FIX_NOT_FOUND
    if not silent:
        _fix_display_message_prelude(
            config[ConfigKeys.DATABASES][ConfigKeys.db_name],
            config[ConfigKeys.DATABASES][ConfigKeys.working_dir],
            data_file
        )

    im = IntegrityManager(config, data_file)
    im.parse_and_fix(silent)
    if not silent:
        _fix_display_message_epilogue()
    if not integrity_cancel:
        integrity_check(_cfg, "None", "None", _sil)



import os
import shutil

from sqlalchemy.orm import mapper

from BioMetaDB import BaseData, Record
from BioMetaDB.Config.config_manager import ConfigKeys
from BioMetaDB.Config.config_manager import ConfigManager
from BioMetaDB.Config.directory_manager import Directories
from BioMetaDB.DBManagers.class_manager import ClassManager
from BioMetaDB.DBOperations.summarize_database import load_table_metadata
from BioMetaDB.Serializers.count_table import CountTable
from BioMetaDB.DBOperations.integrity_check import integrity_check

"""
Script holds functionality for UPDATE, which updates existing/adds new records to database
Handles new column generation and SQL migration 

"""


def _move_project_display_message_prelude(project_name, old_path, new_path):
    """ Summary for moving project

    :param old_path: (str)
    :param new_path: (str)
    :return:
    """
    print("MOVE:\tMove project to new location")
    print(" Old project root directory:\t%s" % old_path.replace(project_name, ""))
    print(" New project root directory:\t%s" % new_path)


def _move_project_display_message_epilogue():
    print("Move complete!", "\n")


def move_project(config_file, path, integrity_cancel, silent):
    """

    :param config_file:
    :param path:
    :param integrity_cancel:
    :param silent:
    :return:
    """
    assert path != 'None', "Path (-p) does not exist, exiting"
    current_path = os.path.abspath(os.path.abspath(os.path.relpath(config_file)))
    path = os.path.abspath(os.path.relpath(path))
    config, config_file = ConfigManager.confirm_config_set(config_file)
    old_path = config[ConfigKeys.DATABASES][ConfigKeys.working_dir]
    assert os.path.dirname(old_path) != os.path.abspath(path), \
        "Project exists in directory, cancelling"
    project_name = config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir]
    if not silent:
        _move_project_display_message_prelude(
            project_name,
            old_path,
            path),
    # Move directory from old location to new location
    shutil.move(current_path, path)
    # Update config file with new location
    abs_path_working_dir = os.path.abspath(os.path.join(path, project_name))
    db_name = project_name
    config[ConfigKeys.DATABASES] = {
        ConfigKeys.db_name: db_name,
        ConfigKeys.working_dir: abs_path_working_dir,
        ConfigKeys.rel_work_dir: db_name,
        ConfigKeys.migrations_dir: os.path.join(abs_path_working_dir, Directories.MIGRATIONS),
        ConfigKeys.config_dir: os.path.join(abs_path_working_dir, Directories.CONFIG),
        ConfigKeys.db_dir: os.path.join(abs_path_working_dir, Directories.DATABASE),
        ConfigKeys.rel_db_dir: os.path.join(db_name, Directories.DATABASE),
    }
    with open(os.path.join(path, project_name, Directories.CONFIG, os.path.basename(config_file)), "w") as W:
        config.write(W)
    # Update location for each record in each table
    tables_in_database = config[ConfigKeys.TABLES_TO_DB].keys()
    engine = BaseData.get_engine(
        os.path.join(config[ConfigKeys.DATABASES][ConfigKeys.working_dir], Directories.DATABASE),
        config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir] + ".db"
    )
    sess = BaseData.get_session_from_engine(engine)
    for tbl_name in tables_in_database:

        TableClass = ClassManager.get_class_orm(tbl_name, engine)
        UserClass = type(tbl_name, (Record,), {})
        mapper(UserClass, TableClass)
        for record in sess.query(UserClass).all():
            record.location = os.path.join(abs_path_working_dir, Directories.DATABASE, tbl_name)
        sess.commit()
        if not integrity_cancel:
            integrity_check(abs_path_working_dir, tbl_name, "None", silent)
    if not silent:
        _move_project_display_message_epilogue()



#!/usr/bin/env python3

import os
import re
from Bio import SeqIO
import itertools as it
from Bio.Seq import Seq
from Bio.Alphabet import IUPAC
from collections import Counter
from BioMetaDB.Accessories.bio_ops import BioOps
from BioMetaDB.Accessories.ops import read_until
from BioMetaDB.Accessories.arg_parse import ArgParse
from BioMetaDB.Indexers.index_mapper import IndexMapper
from BioMetaDB.Indexers.index_mapper import IndexCreator
from BioMetaDB.Indexers.index_mapper import IndexExtensions
from BioMetaDB.Accessories.program_caller import ProgramCaller
from BioMetaDB.Exceptions.get_exceptions import SequenceIdNotFoundError
from BioMetaDB.Exceptions.get_exceptions import ImproperFormatIndexFileError

"""
Script handles file IO operations

"""


def get_seq_from_file(file_name, seq_id):
    """ Locate sequence by ID/description (regex supported) and save to file

    :param file_name: (str)	User-passed name of file
    :param seq_id: (str)	User-passed sequence id to get from file
    """
    file_type = BioOps.get_type(file_name)
    value_to_return = set()
    seq_num = None
    data = {}
    try:
        data = {record.id: record for record in BioOps.parse_large(file_name, file_type)}
        value_to_return = data[seq_id]
    # If value is not found in parsed file, try regex
    except KeyError:
        seq_num = 0
        possible_seqs = list(data.keys())
        possible_descriptions = {val.id: val.description for val in data.values()}
        is_found = False
        # Try each sequence to see if matches regex
        for seq in possible_seqs:
            found = re.search(seq_id, seq)
            if found:
                value_to_return.add(data[seq])
                is_found = True
                seq_num += 1
        # Also check description to see for regex matches
        for _id, desc in possible_descriptions.items():
            found = re.search(seq_id, desc)
            if found:
                value_to_return.add(data[_id])
                is_found = True
                seq_num += 1
        if not is_found:
            print("Could not locate %s in %s" % (seq_id, file_name))
            raise SequenceIdNotFoundError(seq_id)
    file_name = os.path.splitext(file_name)[0]
    # Write found sequence to file
    out_file = "{}.{}".format(seq_id, file_name)
    W = open(out_file, "w")
    SeqIO.write(value_to_return, out_file, file_type)
    W.close()
    if not seq_num:
        print(" Sequence {} copied from file {} to {}".format(seq_id, file_name, out_file))
    else:
        print("{} sequence(s) copied from file {} to {}".format(seq_num, file_name, out_file))


def get_seqs_from_file(file_name, number):
    """ Retrieve a number of sequences from a file

    :param file_name: (str)	Name of file to parse
    :param number: (str)	User-passed number, list or numbers, or range to get
    """
    indices_to_get = _translate_number_string_to_interval(number)
    to_return = []
    data_type = BioOps.get_type(file_name)
    data = BioOps.parse_large(file_name, data_type)
    # Build list of values to return based on user-passed indices (list, range, or value)
    for index in indices_to_get:
        to_return.append(data[index])
    # Write found sequence to file
    out_file = "selected.{}".format(file_name)
    W = open(out_file, "w")
    SeqIO.write(to_return, out_file, data_type)
    W.close()
    print(" {} sequences copied from file {} to {}".format(len(indices_to_get), file_name, out_file))


def _translate_number_string_to_interval(number):
    """ Parses user-passed string for indices to get.
    Supported:
        0-4		Returns indices [0,1,2,3,4]
        1,3,5	Returns indices [1,3,5]

    :param number: (str)	User-passed string for numbers to get
    :return List[int]:
    """
    # List of values to get
    if "," in number:
        number = number.split(",")
        number = [int(_n) for _n in number]
        return number
    # Range of values to get
    else:
        number = number.strip(" ").split("-")
        number = [int(num) for num in number]
        if len(number) == 2:
            number = [_n for _n in range(number[0], number[1] + 1)]
        return number


def get_by_directory(directory):
    """ Renames all files in directory based on filename

    :param directory: (str)		Name of directory with files to rename
    """
    # Correct passed data type based on extension
    # New directory name and create if not exists
    new_location = os.path.join(directory, "tmp")
    if not os.path.isdir(new_location):
        os.makedirs(new_location)
    # Get all files in user-passed directory
    for file in [f for f in os.listdir(directory) if os.path.isfile(f)]:
        # Split file name from extension, get name of old file
        data_type = BioOps.get_type(file)
        _file, ext = os.path.splitext(file)
        fullpath_old = os.path.join(directory, file)
        # Get all records in file
        records_in_file = BioOps.parse_large(fullpath_old, data_type)
        # Rename ids based on file name, add number to end to distinguish
        num = 0
        for record in records_in_file:
            record.id += "_" + _file + "_" + str(num)
            num += 1
        # Write renamed version of file to .tmp directory
        fullpath_new = os.path.join(new_location, _file + ext)
        with open(fullpath_new, "w") as O:
            SeqIO.write(records_in_file, O, data_type)


def get_by_file(file_name):
    """ Renames sequences in file based on file name

    :param file_name: (str)	Name of file to rename
    """
    # Location to write data
    new_location = file_name + ".tmp"
    # Parse data file
    data_type = BioOps.get_type(file_name)
    records_in_file = BioOps.parse_large(file_name, data_type)
    num = 0
    for record in records_in_file:
        # Rename sequence id based on file name and a number for uniqueness
        record.id = file_name + "_" + str(num)
        num += 1
    # Write sequences to new file
    with open(new_location, "w") as O:
        SeqIO.write(records_in_file, O, data_type)


def get_t_to_u(file_name):
    """ Converts thymine to uracil in line

    :param file_name: (str)	Name of fasta file to amend
    """
    data = []
    with open(file_name, "U") as A:
        for line in A:
            # Only change non-header lines
            if not line.startswith(">"):
                line = line.replace("T", "U")
                line = line.replace("t", "u")
            data.append(line)
    with open(file_name, "w") as W:
        for _dat in data:
            W.write(_dat)


def get_u_to_t(file_name):
    """ Converts uracil to thymine in line

    :param file_name: (str)	Name of fasta file to amend
    """
    data = []
    with open(file_name, "U") as A:
        for line in A:
            # Only change non-header lines
            if not line.startswith(">"):
                line = line.replace("U", "T")
                line = line.replace("u", "t")
            data.append(line)
    with open(file_name, "w") as W:
        for _dat in data:
            W.write(_dat)


def reverse_complement_file(file_name):
    """ Reverse complements fasta sequences in file

    :param file_name: (str)	Name of file
    """
    data_type = BioOps.get_type(file_name)
    records_in_file = BioOps.parse_large(file_name, data_type)
    for record in records_in_file:
        record.reverse_complement()
        record.id += ".revComp"
    SeqIO.write(records_in_file, "tmp." + file_name, data_type)


def remove_ambiguity_from_file(file_name):
    """ Removes N from data file in-line

    :param file_name: (str)	Name of file to edit
    """
    data = []
    data_type = BioOps.get_type(file_name)
    records_in_file = BioOps.parse_large(file_name, data_type)
    for record in records_in_file:
        # Build new sequence
        new_seq = ""
        for val in record.seq:
            # Skip ambiguous character
            if val.upper() not in ("R", "Y", "W", "S", "M", "K", "H", "B", "V", "D", "N"):
                new_seq += val
        # Try DNA sequence type first (more restrictive)
        try:
            record.seq = Seq(new_seq, IUPAC.unambiguous_dna)
        # Default to protein otherwise
        except:
            record.seq = Seq(new_seq, IUPAC.unambiguous_rna)
        data.append(record)
    SeqIO.write(data, file_name, data_type)


def summarize_file(file_name, view):
    """ Outputs character count summary of file

    :param view: (str)          User-passed view value
    :param file_name: (str)		Name of file for which to gather character data
    """
    # Corrected view name and data type
    data_type = BioOps.get_type(file_name)
    view = _view_corrector(view)
    # File to read
    record_metadata = {}
    num_records = 0
    R = open(file_name, "rb")
    # Collect file data
    data = []
    if data_type == "fasta":
        # Read in file by fasta header, removing newlines
        for key, group in it.groupby(R, lambda line: line.decode().startswith(">")):
            data.append([gr.decode().rstrip("\r\n").strip(">") for gr in list(group)])
        # Retain record number
        num_records = len(data) // 2
        # Collect metadata by record
        for i in range(0, len(data), 2):
            record_metadata[data[i][0]] = (Counter(data[i][0]), Counter("".join(data[i + 1])))
    elif data_type == "fastq":
        # Read in file by fastq header, removing newlines
        for key, group in it.groupby(R, lambda line: line.startswith("@")):
            data.append([gr.rstrip("\r\n").strip("@") for gr in list(group)])
        # Retain record number
        num_records = len(data) // 2
        # Collect metadata by record
        # Note indices based on presence on "+" in records
        for i in range(0, len(data), 2):
            plus_loc = data[i + 1].index("+")
            record_metadata[data[i][0]] = (Counter(data[i][0]), Counter("".join(data[i + 1][:plus_loc])),
                                           Counter("".join(data[i + 1][plus_loc + 1:])))
    R.close()
    # Output general summary
    _summary_all(num_records, file_name)
    # Output based on user-passed value
    if view == "s":
        _summary_short(record_metadata)
    elif view == "l":
        _summary_long(record_metadata)


def _view_corrector(view):
    """ Returns s or l depending on value passed by user

    :param view: (str)	User-passed view of summary
    :return str:
    """
    if view == "short" or view == "s" or view == "S":
        view = "s"
    elif view == "long" or view == "l" or view == "L":
        view = "l"
    else:
        print("Summary type not found, default to short")
        view = "s"
    return view


def _summary_all(num, file_name):
    """ Displays name of file and number of records

    :param num: (int)			Number of records found
    :param file_name:	(str)	Name of file
    """
    print("%s summary:" % file_name)
    print("Number of records: %i" % num)


def _summary_short(record_metadata):
    """ Short summary that highlighting all data

    :param record_metadata: (Dict[str, Tuple[Counter, Counter]])	Dict with file data
    """
    all_header_metadata = Counter("")
    all_sequence_metadata = Counter("")
    all_quality_metadata = Counter("")
    for met_tuple in record_metadata.values():
        all_header_metadata += met_tuple[0]
        all_sequence_metadata += met_tuple[1]
        if len(met_tuple) == 3:
            all_quality_metadata += met_tuple[2]
    print("Values found in headers:")
    for k, v in all_header_metadata.items():
        print("# %s: %i" % (k, v))
    print("Values found in sequences:")
    for k, v in all_sequence_metadata.items():
        print("# %s: %i" % (k, v))
    if len(all_quality_metadata.keys()) > 0:
        print("Values found in quality scores:")
        for k, v in all_sequence_metadata.items():
            print("# %s: %i" % (BioOps.calculate_phred([k])[0], v))


def _summary_long(record_metadata):
    """ Print summary that highlights each record's information

    :param record_metadata: (Dict[str, Tuple[Counter, Counter]])	Dict with file data
    """
    sorted_keys = sorted(record_metadata.keys())
    for _id in sorted_keys:
        met_tuple = record_metadata[_id]
        print("Id: %s" % _id)
        print("Values found in header:")
        for k, v in met_tuple[0].items():
            print("# %s: %i" % (k, v))
        print("Values found in sequences:")
        for k, v in met_tuple[1].items():
            print("# %s: %i" % (k, v))
        if len(met_tuple) == 3:
            print("Values found in quality scores:")
            for k, v in met_tuple[2].items():
                print("# %s: %i" % (BioOps.calculate_phred([k])[0], v))


def make_file_index(file_name):
    """ Create index of fastx file sequence ids

    :param file_name: (str)	Name of fasta file to index
    """
    # Use fasta file to create index matching short names with ids
    IndexCreator.create_from_fastx(file_name)


def make_list_index(file_name):
    """ Creates simple index using simple list

    :param file_name: (str) Name of file
    :return:
    """
    IndexCreator.create_from_file(file_name)


def make_dir_index(directory):
    """ Creates an index mapping for all files in a directory

    :param directory: (str) Name of directory
    :return:
    """
    IndexCreator.make_idx_for_directory_file_names(directory)


def rename_files_in_dir_using_index(directory):
    """ Renames all files in directory with index file

    :param directory:
    :return:
    """
    IndexCreator.rename_directory_contents_using_index(directory)


def make_fastx_indexed(file_name):
    """ Rename ids in fastx file based on index file

    :param file_name: (str)	Name of fastx file
    """
    # Load index file for user-passed file
    im = IndexMapper(file_name)
    # Write amended fastx file
    IndexCreator.rewrite_ids_in_fastx(im, file_name)


def make_tsv_indexed(file_name, index_file, comment, skip):
    """ Replace sequence ids (first column) in .tsv file with index file ids

    :param file_name: (str)		Path to .tsv file to index
    :param index_file: (str)			Path to index (.imidx) file
    :param comment:	(List[str])	List of comment delimiters to exclude in parsing
    :param skip: (str)			Number of lines at head to exclude in parsing
    """
    im = IndexMapper(index_file)
    # Collect data in list
    data = []
    # Gather comment identifiers
    comments = comment.split(",")
    with open(file_name, "rb") as R:
        # Skip user-determined header lines, store in list
        for i in range(int(skip)):
            data.append(R.readline().decode())
        for line in R:
            line = line.decode()
            # Store comments in data
            if line[0: max([len(_comment) for _comment in comments])] in comments:
                data.append(line)
            # Index remaining items based on file
            else:
                _id = read_until(line, "\t")
                # Replace with indexed value
                line = line.replace(_id, im.get(_id))
                # Store in data
                data.append(line)
    # Rewrite to file-name with extension for indexed file
    with open(file_name + IndexExtensions.IDX_TSV_FILE, "w") as W:
        for line in data:
            W.write(line)


if __name__ == '__main__':
    args_list = [
        [["program"],
         {"help": "Select program"}],
        [["-f", "--file_name"],
         {"help": "Data file for reading/editing"}],
        [["-x", "--index_file"],
         {"help": "Index file with complete and short ids"}],
        [["-d", "--directory"],
         {"help": "Name of directory"}],
        [["-i", "--seq_id"],
         {"help": "Sequence ID to grab (regex)"}],
        [["-v", "--view"],
         {"help": "See in (l)ong or (s)hort format"}],
        [["-n", "--number"],
         {"help": "Number of sequences, inclusive range (index 0), or comma-separated list"}],
        [["-s", "--skip"],
         {"help": "Number of lines for which to skip indexing operation (default 0)", "default": "0"}],
        [["-c", "--comment"],
         {"help": "Comma-separated comment values indicating lines to retain but not parse", "default": "#"}],
    ]
    programs = {
        "ID": get_seq_from_file,
        "NUM": get_seqs_from_file,
        "RDIR": get_by_directory,
        "RFILE": get_by_file,
        "TtoU": get_t_to_u,
        "UtoT": get_u_to_t,
        "REVCOMP": reverse_complement_file,
        "REMOVEAMB": remove_ambiguity_from_file,
        "SUMMARIZE": summarize_file,
        "MAKEIDX": make_file_index,
        "MAKEDIRIDX": make_dir_index,
        "MAKELISTIDX": make_list_index,
        "IDXTOFX": make_fastx_indexed,
        "IDXTOTSV": make_tsv_indexed,
        "IDXTODIR": rename_files_in_dir_using_index,
    }
    flags = {
        "ID": ("file_name", "seq_id"),
        "NUM": ("file_name", "number"),
        "RDIR": ("directory",),
        "RFILE": ("file_name",),
        "TtoU": ("file_name",),
        "UtoT": ("file_name",),
        "REVCOMP": ("file_name",),
        "REMOVEAMB": ("file_name",),
        "SUMMARIZE": ("file_name", "view"),
        "MAKEIDX": ("file_name",),
        "MAKEDIRIDX": ("directory",),
        "MAKELISTIDX": ("file_name",),
        "IDXTOFX": ("file_name",),
        "IDXTOTSV": ("file_name", "index_file", "comment", "skip"),
        "IDXTODIR": ("directory",)
    }
    errors = {
        FileNotFoundError: "Provided filename does not exist",
        SequenceIdNotFoundError: "Sequence ID not found in file",
        ValueError: "File type cannot be parsed by BioPython",
        IndexError: "Sequence number or range exceeds data in file",
        IOError: "File path does not exist",
        TypeError: "Incorrect sequence type in file",
        ImproperFormatIndexFileError: "File is not in .imidx format",
    }
    _help = {
        "ID": "Program for getting a sequence from a file regex-matching request",
        "NUM": "Program for getting a sequence from a file by number, range, or list",
        "RDIR": "Program for renaming entire directory contents based on filenames",
        "RFILE": "Program for renaming sequences in a file based on filename",
        "TtoU": "Program converts thymine bases to uracil in file",
        "UtoT": "Program converts uracil bases to thymine in file",
        "REVCOMP": "Program reverse-complements fasta records in file",
        "REMOVEAMB": "Program removes ambiguous letters from sequence",
        "SUMMARIZE": "Program creates simple summary of sequences in file",
        "MAKEIDX": "Program creates index of fastx file",
        "IDXTOFX": "Program converts fasta file to indexed version (or back!)",
        "IDXTOTSV": "Program converts tsv file to indexed version (or back!)",
        "IDXTODIR": "Program renames contents of directory using index values",
        "MAKEDIRIDX": "Create index for names of file in a directory",
        "MAKELISTIDX": "Create index for list file",
    }

    ap = ArgParse(args_list,
                  description=ArgParse.description_builder("get:\tProgram for basic biological file get operations.",
                                                           _help, flags))

    pc = ProgramCaller(programs=programs, flags=flags, errors=errors, _help=_help)

    pc.run(ap.args, debug=False)



import os
import glob
from configparser import ConfigParser
from BioMetaDB.Config.directory_manager import Directories
from BioMetaDB.Exceptions.config_manager_exceptions import TableNameNotFoundError, MultipleTablesFoundError, \
    ConfigFileNotFound
from BioMetaDB.Exceptions.config_manager_exceptions import ConfigAssertString


class Config(ConfigParser):
    """ Placeholder class for calling configparser

    """
    pass


class ConfigKeys:
    """ Wrapper class holds const strings for simple access

    """
    TABLES_TO_DB = "TABLES_TO_DB"
    TABLES_TO_ALIAS = "TABLES_TO_ALIAS"
    working_dir = "working_dir"
    migrations_dir = "migrations_dir"
    config_dir = "config_dir"
    db_dir = "db_dir"
    rel_work_dir = "rel_work_dir"
    rel_db_dir = "rel_db_dir"
    rel_classes_dir = "rel_classes_dir"
    DATABASES = "DATABASES"
    db_name = "db_name"


class ConfigManager:

    def __init__(self, config, table_name):
        """ Serializes output from configparser object for ease of access

        :param config: (Config)     Config object using project config file
        :param table_name: (str)    Name of table
        """
        if table_name not in config[ConfigKeys.TABLES_TO_DB].keys():
            raise TableNameNotFoundError(table_name)
        self.config = config
        self.table_name = table_name
        self.db_name = config[ConfigKeys.TABLES_TO_DB][table_name]
        self.working_dir = config[ConfigKeys.DATABASES][ConfigKeys.working_dir]
        self.db_file = config[ConfigKeys.TABLES_TO_DB][table_name] + ".db"
        self.db_dir = os.path.join(config[ConfigKeys.DATABASES][ConfigKeys.working_dir], Directories.DATABASE)
        self.table_dir = os.path.join(config[ConfigKeys.DATABASES][ConfigKeys.db_dir], table_name)
        self.classes_file = os.path.join(config[ConfigKeys.DATABASES][ConfigKeys.working_dir], "classes", table_name + ".json")
        self.rel_work_dir = config[ConfigKeys.DATABASES][ConfigKeys.rel_work_dir]
        self.rel_db_dir = config[ConfigKeys.DATABASES][ConfigKeys.rel_db_dir]
        self.rel_classes_dir = config[table_name][ConfigKeys.rel_classes_dir]
        self.migrations_dir = config[ConfigKeys.DATABASES][ConfigKeys.migrations_dir]
        self.config_dir = config[ConfigKeys.DATABASES][ConfigKeys.config_dir]

    def update_config_file(self, table_name):
        """ Creates updated table info in config file, writes new file

        :param table_name:
        :return:
        """
        self.config[table_name] = {
            ConfigKeys.rel_classes_dir: os.path.join(self.working_dir, Directories.CLASSES),
        }
        self.config[ConfigKeys.TABLES_TO_DB][table_name] = self.db_name
        with open(os.path.join(os.path.join(self.working_dir, Directories.CONFIG), self.db_name + ".ini"), "w") as W:
            self.config.write(W)

    def remove_table_from_config_file(self, table_name):
        """ Removes table info from config file and rewrites

        :param table_name:
        :return:
        """
        del self.config[table_name]
        del self.config[ConfigKeys.TABLES_TO_DB][table_name]
        with open(os.path.join(os.path.join(self.working_dir, Directories.CONFIG), self.db_name + ".ini"), "w") as W:
            self.config.write(W)

    @staticmethod
    def confirm_config_set(config_file):
        """ Static method to confirm that the config file was passed and exists

        :param config_file:
        :return:
        """
        if config_file == 'None' or config_file is None:
            config_file = glob.glob(os.path.join(os.getcwd(), "*/config/*.ini"))
        else:
            config_file = glob.glob(os.path.join(config_file, "config/*.ini"))
        num_files = len(config_file)
        if num_files == 1:
            config_file = config_file[0]
        elif num_files > 1:
            raise MultipleTablesFoundError
        elif num_files == 0:
            raise ConfigFileNotFound
        config = Config()
        config.read(config_file)
        return config, config_file

    @staticmethod
    def get_name_by_alias(alias, config):
        for key in config[ConfigKeys.TABLES_TO_ALIAS].keys():
            key = key.split("|")
            if key[0] == alias:
                return key[1]
        return None
